///////////////////////////////////////////////////////////////////////////////
//
//  Modsaver - Wharf module that runs a screen saver inside it.
//             Original copyright follows.
//
//  Written by Andy Rudnitsky (Netwalker), rudnitsky@voyager.net
//
//  6/16/98 - Origination
//  6/19/98 - Fix to handle WM_SETTINGCHANGE messages
//
///////////////////////////////////////////////////////////////////////////////

/*

  This is a part of the LiteStep Shell Source code.

  Copyright (C) 1997-98 Francis Gastellu
                    aka Lone Runner/Aegis

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <windows.h>
#include <assert.h>
#include <stdio.h>

#include "..\..\..\SDK\lssdk.h"


// Window callback shared by our two windows
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

// prototypes for internal routines
static void TerminateScreenSaverProcess();
static BOOL StartScreenSaverProcess();
static BOOL GetCurrentScreenSaver( LPSTR lpszCurrentScreenSaver );
static BOOL StartScreenSaverInWindow( LPSTR lpszScreenSaver, HWND hWnd );

// globals
static char            gszAppName[]   = "ModSaver"; // Name of Application, Window class...
static HWND            ghMainWnd      = NULL;       // main window handle
static HWND            ghInvisibleWnd = NULL;       // non-child window
static HWND            gParent        = NULL;       // gParent window
static wharfDataType   gWharfData;                  // Setup data passed to the DLL
static int             gWndSize;                    // Size of the wharf client
static HANDLE          ghSaverProcess = NULL;       // Screen saver process handle

// Max length for command line string
#define SAVER_CMD_LINE_MAX      512

///////////////////////////////////////////////////////////////////////////////
//
// Wharf Dll initialization entry point
//
///////////////////////////////////////////////////////////////////////////////
__declspec( dllexport ) int initWharfModule( HWND ParentWnd, HINSTANCE dllInst, wharfDataType * wd )
{
    WNDCLASS    ModSaverClass;
    BOOL        Ret;

    // Save gParent window
    gParent = ParentWnd;    

    // Duplicate gWharfData since the one we get will be destroyed
    memcpy( &gWharfData, wd, sizeof(wharfDataType) );
   
    // Calculate Wharf module size
    // @@@@LSDEV - we need #defines for these magic numbers in API
    gWndSize = 64 - gWharfData.borderSize * 2;     


    // Register the Window class
    memset( &ModSaverClass, 0, sizeof(WNDCLASS) );
    ModSaverClass.lpfnWndProc   = WndProc;     // our window procedure
    ModSaverClass.hInstance     = dllInst;     // hInstance of DLL
    ModSaverClass.lpszClassName = gszAppName;  // our window class name

    if ( ! RegisterClass( &ModSaverClass ) ) 
    {
        MessageBox( gParent, "Error registering window class", gszAppName, MB_OK );
        return 1;
    }

    // Now create the two windows we will use
    // This is the window that will show the screen saver
    ghMainWnd = CreateWindowEx(
        0,                                            // exstyles 
        gszAppName,                                   // our window class name
        gszAppName,                                   // use description for a window title
        WS_CHILD,                                     // window style
        gWharfData.borderSize, gWharfData.borderSize, // position 
        gWndSize,gWndSize,                            // width & height of window
        gParent,                                      // gParent window (litestep wharf window)
        NULL,                                         // no menu
        dllInst,                                      // hInstance of DLL
        0);                                           // no window creation data

    if ( ! ghMainWnd ) 
    {						   
        MessageBox( gParent, "Error creating module window", gszAppName, MB_OK );
        return 1;
    }

    // This window is NOT a child, so it will get the WM_SETTINGCHANGE
    // whenever the screen saver changes.  Note that both windows are
    // the same class so they share the same window procedure
    ghInvisibleWnd = CreateWindowEx(
        0,                                            // exstyles 
        gszAppName,                                   // our window class name
        gszAppName,                                   // use description for a window title
        WS_DISABLED,                                  // window style
        0, 0,                                         // position 
        0, 0,                                         // width & height of window
        gParent,                                      // gParent window (litestep wharf window)
        NULL,                                         // no menu
        dllInst,                                      // hInstance of DLL
        0);                                           // no window creation data
    
    if ( ! ghInvisibleWnd ) 
    {
        // Clean up
        DestroyWindow( ghMainWnd );                
        MessageBox( gParent, "Error creating module window", gszAppName, MB_OK );
        return 1;
    }

//    // Set normal cursor
//    SetCursor( LoadCursor( NULL, MAKEINTRESOURCE(IDC_ARROW) ) );
	
    // DO NOT REMOVE ! Set magicDWord - required!
    SetWindowLong( ghMainWnd, GWL_USERDATA, magicDWord ); 

    // show the window
    ShowWindow( ghMainWnd, SW_SHOWNORMAL );

    // Hide the message procesing window
    ShowWindow( ghInvisibleWnd, SW_HIDE );

    // Attempt to start the screen saver
    Ret = StartScreenSaverProcess();
    if ( Ret == FALSE ) 
        // Failed for whatever reason
        return 1;
    else
        return 0;
}

///////////////////////////////////////////////////////////////////////////////
//
// Wharf Dll closeup request
//
///////////////////////////////////////////////////////////////////////////////
__declspec( dllexport ) void quitWharfModule( HINSTANCE dllInst )
{
    // Kill off the screen saver
    TerminateScreenSaverProcess();

    DestroyWindow( ghInvisibleWnd );           // delete msg processing window
    DestroyWindow( ghMainWnd );                // delete our display window
    UnregisterClass( gszAppName, dllInst );    // unregister window class
}

///////////////////////////////////////////////////////////////////////////////
//
//  Modsaver window procedure
//
///////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_SETTINGCHANGE:
            // Reload the new screen saver irregardless of what changed
            // It's just easier to do this....
            TerminateScreenSaverProcess();
            StartScreenSaverProcess();
            return 0;

        
        case WM_CREATE:		
            return 0;

        
        case WM_ERASEBKGND: 
            return 0;

        
        case WM_KEYDOWN:    
        case WM_KEYUP:
            // forward keyboard messages to gParent window 
            PostMessage( gParent, message, wParam, lParam );
            return 0;
	    

        // Mouse messages are here to ensure we have a good popup menu behaviour. You may insert
        // your own custom actions
        case WM_RBUTTONUP:
        {
            RECT    r;
            GetWindowRect( hwnd, &r );
            PostMessage( GetParent( GetParent( gParent ) ), 
                         9182, 
                         r.top + (int)HIWORD(lParam), 
                         r.left + (int)LOWORD(lParam) );
        }
        return 0;


        case WM_RBUTTONDOWN:
        case WM_LBUTTONDOWN:
        case WM_MBUTTONDOWN:
            PostMessage( GetParent( GetParent( gParent ) ), 9183, 0, 0);
            return 0;
    }

    // Pass everything else along
    return DefWindowProc( hwnd, message, wParam, lParam );
}

///////////////////////////////////////////////////////////////////////////////
//
//  StartScreenSaverProcess
//
///////////////////////////////////////////////////////////////////////////////
static BOOL StartScreenSaverProcess()
{
    char    szScreenSaver[ SAVER_CMD_LINE_MAX ];
    BOOL    Ret;


    Ret = GetCurrentScreenSaver( szScreenSaver );
    if ( Ret == FALSE )
    {
        // No current saver
        MessageBox( gParent, "No current screen saver", gszAppName, MB_OK );
    }
    else
    {
        Ret = StartScreenSaverInWindow( szScreenSaver, ghMainWnd );
        if ( Ret == FALSE ) 
            MessageBox( gParent, "Error creating screen saver process", gszAppName, MB_OK );
    }

    return Ret;
}

///////////////////////////////////////////////////////////////////////////////
//
//  GetCurrentScreenSaver
//
///////////////////////////////////////////////////////////////////////////////
static BOOL GetCurrentScreenSaver( LPSTR lpszCurrentScreenSaver )
{
    assert( lpszCurrentScreenSaver );

    (void)GetPrivateProfileString(  "boot",                 // Section
                                    "SCRNSAVE.EXE",         // Key
                                    NULL,                   // Default
                                    lpszCurrentScreenSaver, // Destination buffer
                                    SAVER_CMD_LINE_MAX,     // Buffer length
                                    "system.ini" );         // Filename

    if ( *lpszCurrentScreenSaver == '\0' )
        // No screen saver entry
        return FALSE;
    else
    {
        return TRUE;
    }
}

///////////////////////////////////////////////////////////////////////////////
//
//  StartScreenSaverInWindow
//
///////////////////////////////////////////////////////////////////////////////
static BOOL StartScreenSaverInWindow( LPSTR lpszScreenSaver, HWND hWnd )
{
    STARTUPINFO          siStartupInfo;
    PROCESS_INFORMATION  piProcessInfo;
    BOOL                 Status;
    char                 szScreenSaverCommandLine[ SAVER_CMD_LINE_MAX ];
    char                 szCommandArgs[ SAVER_CMD_LINE_MAX ];

    assert( lpszScreenSaver );

    // Note that not all screen savers support preview
    sprintf( szCommandArgs, " /P %d", (int)ghMainWnd );
    if ( strlen( szCommandArgs ) + strlen( szScreenSaverCommandLine ) > (SAVER_CMD_LINE_MAX - 1) )
        // wont't fit
        return FALSE;

    // Build the command line string
    strcpy( szScreenSaverCommandLine, lpszScreenSaver );
    strcat( szScreenSaverCommandLine, szCommandArgs );


    //  Initialize info structures
    memset( &siStartupInfo, 0, sizeof(STARTUPINFO) );
    memset( &piProcessInfo, 0, sizeof(PROCESS_INFORMATION) );

    // Use current process's StartupInfo
    GetStartupInfo( &siStartupInfo );
    
    Status = CreateProcess( NULL,                               // app name
                            szScreenSaverCommandLine,           // command line string
                            NULL,                               // process security attributes
                            NULL,                               // thread security attributes
                            FALSE,                              // handle inheritance flag
                            IDLE_PRIORITY_CLASS,                // creation flags
                            NULL,                               // new environment block
                            NULL,                               // current directory name
                            &siStartupInfo,                     // STARTUPINFO
                            &piProcessInfo );                   // PROCESS_INFORMATION

    if( Status )
    {
       // Save off the process handle for later termination
       ghSaverProcess = piProcessInfo.hProcess;
       return TRUE;
    }
    else   
    {
       // Not able to start the saver
       ghSaverProcess = NULL;
       return FALSE;
    }

}

///////////////////////////////////////////////////////////////////////////////
//
//  TerminateScreenSaver
//
///////////////////////////////////////////////////////////////////////////////
static void TerminateScreenSaverProcess()
{
    if( ghSaverProcess )
        TerminateProcess( ghSaverProcess, 0 );

    ghSaverProcess = NULL;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
